import setuptools

setuptools.setup(
    name="Jong",
    version="1.0.5",
    license='MIT',
    author="Syniez",
    author_email="js990204@naver.com",
    description="https://github.com/Syniez/Syniez_module.git",
    long_description=open('README.md').read(),
    url="https://github.com/Syniez/Syniez_module.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.6',
)